# cython: profile=True
# distutils: define_macros=CYTHON_TRACE_NOGIL=1

# Author: Tom.Bochynek@gmail.com

from __future__ import division
cimport numpy as np
cimport cython
import numpy as np
import time
import csv
import datetime
import random
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

###
# GlOBALS
###

# I use terminology of classical Prisoner's Dilemma game.

cdef float TEMPTATION = 1.22313
cdef float REWARD = 1.19255
cdef float SUCKER = 0.975
cdef float PENALTY = 1


###
# FUNCTIONS
###

def save_plot(results, save_string):
    """
    saves final plot of avg investment of generations
    """
    plt.clf()
    plt.plot(results)
    plot_save_path = 'results/'
    plot_save_name = save_string + '.pdf'
    plt.savefig(plot_save_path + plot_save_name)


def write_results(results, save_string):
    """
    wites output to file
    """
    csv_path = 'results/'
    csv_name = save_string + '.csv'

    with open(csv_path+csv_name, "w") as output:
        writer = csv.writer(output, lineterminator='\n')
        for val in results:
            writer.writerow([val])


def write_log(start_time, end_time, grid_length, neighbour_distance, mutation_chance, init_min_inv, init_max_inv):
    """
    expands running log on cluster to keep track of experiments conducted
    """
    log_entry = [start_time, end_time, grid_length, neighbour_distance, mutation_chance, init_min_inv, init_max_inv]
    with open('results/log.csv', 'a') as f:
        writer = csv.writer(f)
        writer.writerow(log_entry)


def dump_progress_log(save_string, current_progress, gen):
    """
    at specified intervals, saves progress to separate folder
    """
    dump_time = datetime.datetime.now().strftime("%d-%m-%Y_%H-%M")
    dump_path = 'running_dumps/'
    dump_csv_name = 'dump_' + save_string + '__gen' + str(gen) + '_t' + dump_time + '.csv'

    # writing csv
    with open(dump_path + dump_csv_name, "w") as output:
        writer = csv.writer(output, lineterminator='\n')
        for val in current_progress:
            writer.writerow([val])

    # saving plot
    plt.clf()
    plt.plot(current_progress)
    dump_plot_name = 'dump_' + save_string + '__gen' + str(gen) + '_t' + dump_time + '.pdf'
    plt.savefig(dump_path + dump_plot_name)


cdef inline float get_player_payoff(float a, float b):
    """
    calculates payoff value form continuous payoff function.
    :param a: investment of player a
    :param b: investment of player b
    :return: payoff for player a
    """
    cdef float payoff

    payoff = (a * b * REWARD) + \
             (a * (1 - b) * SUCKER) + \
             ((1 - a) * b * TEMPTATION) + \
             ((1 - a) * (1 - b) * PENALTY)

    return payoff


cdef list play_games(int grid_length, int max_generation, float mutation_chance, float move_chance, int neighbour_distance, float init_min_inv, float init_max_inv, save_string):
    # setting up matrices
    cdef np.ndarray current_inv_matrix = np.random.uniform(init_min_inv, init_max_inv,
                                                                          [grid_length, grid_length])  # make kernel
    cdef int kernel_side_length = 2 * neighbour_distance + 1

    # number of interactions for each player in the game:
    cdef int num_interactions = 2 * ((kernel_side_length*kernel_side_length) - 1)

    # list of average investments per round
    cdef list avg_investments = []

    # other declarations
    cdef list mutant_coord_list, mover_coord_list, grid_range, kernel_range_x, kernel_range_y
    cdef int generation, s, t, x, y, rand_x, rand_y
    cdef int min_crop_x, max_crop_x, min_crop_y, max_crop_y
    cdef float player_pay_off_collection, curr_player_pay_off, holder
    cdef int opponent_x, opponent_y, win_y, win_x
    cdef int mutant_x, mutant_y, mover_x, mover_y
    cdef float curr_play_inv, curr_opp_inv, curr_opp_payoff, mut_rand, move_rand
    cdef float mean, var, mutated_strategy
    cdef np.ndarray payoff_matrix, next_inv_matrix


# generation loop
    generation = 0

    while generation < max_generation:
        generation += 1
        # resetting current generation payoff counter
        payoff_matrix = np.zeros(shape=(grid_length, grid_length), dtype = np.float)

        # resetting list holding future investment
        next_inv_matrix = np.zeros(shape=(grid_length, grid_length), dtype=np.float)

        # list for gathering mutants. this is done in the strategy update loop; saves another full iteration over array
        mutant_coord_list = []

        # list for gathering movers. as above.
        mover_coord_list = []

        grid_range = range(grid_length)

        # plays game for all players. updates player and opponent payoffs.
        for x in grid_range:
            for y in grid_range:
                # determine coordinates of moving window
                min_crop_x = int(x - (kernel_side_length - 1) / 2)
                max_crop_x = int(x + (kernel_side_length - 1) / 2 + 1)
                min_crop_y = int(y - (kernel_side_length - 1) / 2)
                max_crop_y = int(y + (kernel_side_length - 1) / 2 + 1)

                # current player investment:
                curr_play_inv = current_inv_matrix[x][y]
                player_pay_off_collection = 0

                # iterates over each opponent; calculate payoffs
                kernel_range_x = range(min_crop_x, max_crop_x)
                kernel_range_y = range(min_crop_y, max_crop_y)

                for s in kernel_range_x:
                    for t in kernel_range_y:

                        # wrap
                        opponent_x = s % grid_length
                        opponent_y = t % grid_length

                        # catch games with self
                        if opponent_x == x and opponent_y == y:
                            continue

                        # retrieve investments
                        curr_opp_inv = current_inv_matrix[opponent_x][opponent_y]

                        # calculate pay_offs
                        curr_player_pay_off = get_player_payoff(curr_play_inv, curr_opp_inv)
                        player_pay_off_collection += curr_player_pay_off
                        curr_opp_payoff = get_player_payoff(curr_opp_inv, curr_play_inv)
                        payoff_matrix[opponent_x][opponent_y] += curr_opp_payoff

                payoff_matrix[x][y] += player_pay_off_collection

        # average payoffs
        payoff_matrix /= num_interactions


        # decide on investment for next round
        for x in grid_range:
            for y in grid_range:
                # determine coordinates of moving window
                min_crop_x = int(x - (kernel_side_length - 1) / 2)
                max_crop_x = int(x + (kernel_side_length - 1) / 2 + 1)
                min_crop_y = int(y - (kernel_side_length - 1) / 2)
                max_crop_y = int(y + (kernel_side_length - 1) / 2 + 1)

                fitness_dict = {}

                # identify if player mutates at end of round
                mut_rand = random.random()
                if mut_rand <= mutation_chance:
                    mutant_coord_list.append((x, y))

                # identify if player mutates at end of round
                move_rand = random.random()
                if move_rand <= move_chance:
                    mover_coord_list.append((x, y))

                kernel_range_x = range(min_crop_x, max_crop_x)
                kernel_range_y = range(min_crop_y, max_crop_y)

                # iterates of each opponent; builds dictionary of payoff values
                for s in kernel_range_x:
                    for t in kernel_range_y:
                        # wrap
                        opponent_x = s % grid_length
                        opponent_y = t % grid_length

                        # build dictionary of fitness values for all opponents including self
                        fitness_dict[(opponent_x, opponent_y)] = payoff_matrix[opponent_x][opponent_y]

                # selects dictionary entry based on highest fitness value
                win_x, win_y = max(fitness_dict.iterkeys(), key=(lambda key: fitness_dict[key]))
                next_inv_matrix[x][y] = current_inv_matrix[win_x][win_y]

        # building result list
        avg_investments.append(np.average(current_inv_matrix))

        # mutate player from previously built mutant_coord_list
        for mutant_x, mutant_y in mutant_coord_list:
            mean = current_inv_matrix[mutant_x][mutant_y]
            if mean <= init_min_inv:
                mean = init_min_inv
            var = 0.1 * mean
            mutated_strategy = np.random.normal(mean, var)
            if mutated_strategy < init_min_inv:
                mutated_strategy=init_min_inv

            # print 'mut', mutated_strategy
            if mutated_strategy > 1:
                mutated_strategy = 1

            next_inv_matrix[mutant_x][mutant_y] = mutated_strategy

        # update next round's investment
        current_inv_matrix = next_inv_matrix

        # move players gathered previously in mover_coord_list
        for mover_x, mover_y in mover_coord_list:
            # choose random player to switch with
            rand_x = random.randint(0,grid_length-1)
            rand_y = random.randint(0,grid_length-1)

            # stash occupant investment , replace with mover, move replaced player into mover origin
            holder = current_inv_matrix[rand_x][rand_y]
            current_inv_matrix[rand_x][rand_y] = current_inv_matrix[mover_x][mover_y]
            current_inv_matrix[mover_x][mover_y] = holder


        if generation % 50000 == 0:

    # returns average list
    return avg_investments


def run_simulation(int grid_length, int max_generation, float mutation_chance,float move_chance, int neighbour_distance, float init_min_inv, float init_max_inv):
    start_time = datetime.datetime.now().strftime("%d-%m-%Y_%H-%M")
    save_string = 't1-5_pMove' + str(move_chance)  + '_pMut' + str(mutation_chance) + '_t' + start_time

    # playing games
    results = play_games(grid_length, max_generation, mutation_chance, move_chance, neighbour_distance, init_min_inv,
                         init_max_inv, save_string)

    # write outputs to file
    write_results(results, save_string)

    # save plot
    save_plot(results, save_string)

    # conclude
    end_time = datetime.datetime.now().strftime("%d-%m-%Y_%H-%M")
    write_log(start_time, end_time, grid_length, neighbour_distance, mutation_chance, init_min_inv, init_max_inv)

